@@ -1,81 +1,23 @@
-"""
-    Twython is a library for Python that wraps the Twitter API.
-    It aims to abstract away all the API endpoints, so that additions to the library
-    and/or the Twitter API won't cause any overall problems.
-
-    Questions, comments? ryan@venodesigns.net
-"""
-
-__author__ = "Ryan McGrath <ryan@venodesigns.net>"
-__version__ = "2.7.3"
-
-import urllib
 import re
 import warnings
 
 import requests
 from requests_oauthlib import OAuth1
 
-try:
-    from urlparse import parse_qsl
-except ImportError:
-    from cgi import parse_qsl
+from . import __version__
+from .advisory import TwythonDeprecationWarning
+from .compat import json, urlencode, parse_qsl, quote_plus
+from .endpoints import api_table
+from .exceptions import TwythonError, TwythonAuthError, TwythonRateLimitError
 
-# Twython maps keyword based arguments to Twitter API endpoints. The endpoints
-# table is a file with a dictionary of every API endpoint that Twython supports.
-from twitter_endpoints import base_url, api_table, twitter_http_status_codes
-
-try:
-    import simplejson as json
-except ImportError:
-    import json
-
-
-class TwythonError(Exception):
-    """
-        Generic error class, catch-all for most Twython issues.
-        Special cases are handled by TwythonAPILimit and TwythonAuthError.
-
-        Note: To use these, the syntax has changed as of Twython 1.3. To catch these,
-        you need to explicitly import them into your code, e.g:
-
-        from twython import TwythonError, TwythonRateLimitError, TwythonAuthError
-    """
-    def __init__(self, msg, error_code=None, retry_after=None):
-        self.error_code = error_code
-
-        if error_code is not None and error_code in twitter_http_status_codes:
-            msg = 'Twitter API returned a %s (%s), %s' % (error_code,
-                                                          twitter_http_status_codes[error_code][0],
-                                                          msg)
-
-        super(TwythonError, self).__init__(msg)
-
-    @property
-    def msg(self):
-        return self.args[0]
-
-
-class TwythonAuthError(TwythonError):
-    """ Raised when you try to access a protected resource and it fails due to
-        some issue with your authentication.
-    """
-    pass
-
-
-class TwythonRateLimitError(TwythonError):
-    """ Raised when you've hit a rate limit.
-        retry_wait_seconds is the number of seconds to wait before trying again.
-    """
-    def __init__(self, msg, error_code, retry_after=None):
-        if isinstance(retry_after, int):
-            msg = '%s (Retry after %d seconds)' % (msg, retry_after)
-        TwythonError.__init__(self, msg, error_code=error_code)
+warnings.simplefilter('always', TwythonDeprecationWarning)  # For Python 2.7 >
 
 
 class Twython(object):
-    def __init__(self, app_key=None, app_secret=None, oauth_token=None, oauth_token_secret=None,
-                 headers=None, callback_url=None, twitter_token=None, twitter_secret=None, proxies=None, version='1.1'):
+    def __init__(self, app_key=None, app_secret=None, oauth_token=None,
+                 oauth_token_secret=None, headers=None, proxies=None,
+                 version='1.1', callback_url=None, ssl_verify=True,
+                 twitter_token=None, twitter_secret=None):
         """Instantiates an instance of Twython. Takes optional parameters for authentication and such (see below).
 
             :param app_key: (optional) Your applications key
@@ -85,48 +27,59 @@ class Twython(object):
             :param headers: (optional) Custom headers to send along with the request
             :param callback_url: (optional) If set, will overwrite the callback url set in your application
             :param proxies: (optional) A dictionary of proxies, for example {"http":"proxy.example.org:8080", "https":"proxy.example.org:8081"}.
+            :param ssl_verify: (optional) Turns off ssl verification when False. Useful if you have development server issues.
         """
 
-        # Needed for hitting that there API.
+        # API urls, OAuth urls and API version; needed for hitting that there API.
         self.api_version = version
         self.api_url = 'https://api.twitter.com/%s'
         self.request_token_url = self.api_url % 'oauth/request_token'
         self.access_token_url = self.api_url % 'oauth/access_token'
         self.authenticate_url = self.api_url % 'oauth/authenticate'
 
-        # Enforce unicode on keys and secrets
-        self.app_key = app_key and unicode(app_key) or twitter_token and unicode(twitter_token)
-        self.app_secret = app_key and unicode(app_secret) or twitter_secret and unicode(twitter_secret)
-
-        self.oauth_token = oauth_token and u'%s' % oauth_token
-        self.oauth_token_secret = oauth_token_secret and u'%s' % oauth_token_secret
+        self.app_key = app_key or twitter_token
+        self.app_secret = app_secret or twitter_secret
+        self.oauth_token = oauth_token
+        self.oauth_token_secret = oauth_token_secret
 
         self.callback_url = callback_url
 
-        # If there's headers, set them, otherwise be an embarassing parent for their own good.
-        self.headers = headers or {'User-Agent': 'Twython v' + __version__}
+        if twitter_token or twitter_secret:
+            warnings.warn(
+                'Instead of twitter_token or twitter_secret, please use app_key or app_secret (respectively).',
+                TwythonDeprecationWarning,
+                stacklevel=2
+            )
 
-        # Allow for unauthenticated requests
-        self.client = requests.Session()
-        self.client.proxies = proxies
+        if callback_url:
+            warnings.warn(
+                'Please pass callback_url to the get_authentication_tokens method rather than Twython.__init__',
+                TwythonDeprecationWarning,
+                stacklevel=2
+            )
+
+        self.headers = {'User-Agent': 'Twython v' + __version__}
+        if headers:
+            self.headers.update(headers)
+
+        # Generate OAuth authentication object for the request
+        # If no keys/tokens are passed to __init__, self.auth=None allows for
+        # unauthenticated requests, although I think all v1.1 requests need auth
         self.auth = None
+        if self.app_key is not None and self.app_secret is not None and \
+           self.oauth_token is None and self.oauth_token_secret is None:
+            self.auth = OAuth1(self.app_key, self.app_secret)
 
         if self.app_key is not None and self.app_secret is not None and \
-        self.oauth_token is None and self.oauth_token_secret is None:
+           self.oauth_token is not None and self.oauth_token_secret is not None:
             self.auth = OAuth1(self.app_key, self.app_secret,
-                               signature_type='auth_header')
+                               self.oauth_token, self.oauth_token_secret)
 
-        if self.app_key is not None and self.app_secret is not None and \
-        self.oauth_token is not None and self.oauth_token_secret is not None:
-            self.auth = OAuth1(self.app_key, self.app_secret,
-                               self.oauth_token, self.oauth_token_secret,
-                               signature_type='auth_header')
-
-        if self.auth is not None:
-            self.client = requests.Session()
-            self.client.headers = self.headers
-            self.client.auth = self.auth
-            self.client.proxies = proxies
+        self.client = requests.Session()
+        self.client.headers = self.headers
+        self.client.proxies = proxies
+        self.client.auth = self.auth
+        self.client.verify = ssl_verify
 
         # register available funcs to allow listing name when debugging.
         def setFunc(key):
@@ -142,8 +95,8 @@ class Twython(object):
         fn = api_table[api_call]
         url = re.sub(
             '\{\{(?P<m>[a-zA-Z_]+)\}\}',
-            lambda m: "%s" % kwargs.get(m.group(1), self.api_version),
-            base_url + fn['url']
+            lambda m: "%s" % kwargs.get(m.group(1)),
+            self.api_url % self.api_version + fn['url']
         )
 
         content = self._request(url, method=fn['method'], params=kwargs)
@@ -155,15 +108,7 @@ class Twython(object):
         code twice, right? ;)
         '''
         method = method.lower()
-        if not method in ('get', 'post'):
-            raise TwythonError('Method must be of GET or POST')
-
         params = params or {}
-        # requests doesn't like items that can't be converted to unicode,
-        # so let's be nice and do that for the user
-        for k, v in params.items():
-            if isinstance(v, (int, bool)):
-                params[k] = u'%s' % v
 
         func = getattr(self.client, method)
         if method == 'get':
@@ -217,7 +162,7 @@ class Twython(object):
                                 error_code=response.status_code,
                                 retry_after=response.headers.get('retry-after'))
 
-        # if we have a json error here, then it's not an official TwitterAPI error
+        # if we have a json error here, then it's not an official Twitter API error
         if json_error and not response.status_code in (200, 201, 202):
             raise TwythonError('Response was not valid JSON, unable to decode.')
 
@@ -258,10 +203,10 @@ class Twython(object):
             This will return None if the header is not present
 
             Most useful for the following header information:
-                x-ratelimit-limit
-                x-ratelimit-remaining
-                x-ratelimit-class
-                x-ratelimit-reset
+                x-rate-limit-limit
+                x-rate-limit-remaining
+                x-rate-limit-class
+                x-rate-limit-reset
         """
         if self._last_call is None:
             raise TwythonError('This function must be called after an API call.  It delivers header information.')
@@ -269,23 +214,23 @@ class Twython(object):
             return self._last_call['headers'][header]
         return self._last_call
 
-    def get_authentication_tokens(self, force_login=False, screen_name=''):
-        """Returns an authorization URL for a user to hit.
+    def get_authentication_tokens(self, callback_url=None, force_login=False, screen_name=''):
+        """Returns a dict including an authorization URL (auth_url) to direct a user to
 
+            :param callback_url: (optional.. for now) Url the user is returned to after they authorize your app
             :param force_login: (optional) Forces the user to enter their credentials to ensure the correct users account is authorized.
             :param app_secret: (optional) If forced_login is set OR user is not currently logged in, Prefills the username input box of the OAuth login screen with the given value
         """
-        request_args = {}
-        if self.callback_url:
-            request_args['oauth_callback'] = self.callback_url
-
+        callback_url = callback_url or self.callback_url
+        request_args = {'oauth_callback': callback_url}
         response = self.client.get(self.request_token_url, params=request_args)
+
         if response.status_code == 401:
             raise TwythonAuthError(response.content, error_code=response.status_code)
         elif response.status_code != 200:
             raise TwythonError(response.content, error_code=response.status_code)
 
-        request_tokens = dict(parse_qsl(response.content))
+        request_tokens = dict(parse_qsl(response.content.decode('utf-8')))
         if not request_tokens:
             raise TwythonError('Unable to decode request tokens.')
 
@@ -302,18 +247,20 @@ class Twython(object):
             })
 
         # Use old-style callback argument if server didn't accept new-style
-        if self.callback_url and not oauth_callback_confirmed:
+        if callback_url and not oauth_callback_confirmed:
             auth_url_params['oauth_callback'] = self.callback_url
 
-        request_tokens['auth_url'] = self.authenticate_url + '?' + urllib.urlencode(auth_url_params)
+        request_tokens['auth_url'] = self.authenticate_url + '?' + urlencode(auth_url_params)
 
         return request_tokens
 
     def get_authorized_tokens(self, oauth_verifier):
         """Returns authorized tokens after they go through the auth_url phase.
+
+        :param oauth_verifier: (required) The oauth_verifier retrieved from the callback url querystring
         """
         response = self.client.get(self.access_token_url, params={'oauth_verifier': oauth_verifier})
-        authorized_tokens = dict(parse_qsl(response.content))
+        authorized_tokens = dict(parse_qsl(response.content.decode('utf-8')))
         if not authorized_tokens:
             raise TwythonError('Unable to decode authorized tokens.')
 
@@ -326,7 +273,7 @@ class Twython(object):
     # ------------------------------------------------------------------------------------------------------------------------
 
     @staticmethod
-    def shortenURL(url_to_shorten, shortener='http://is.gd/api.php'):
+    def shortenURL(url_to_shorten, shortener='http://is.gd/create.php'):
         """Shortens url specified by url_to_shorten.
             Note: Twitter automatically shortens all URLs behind their own custom t.co shortener now,
                 but we keep this here for anyone who was previously using it for alternative purposes. ;)
@@ -335,11 +282,18 @@ class Twython(object):
             :param shortener: (optional) In case you want to use a different
                               URL shortening service
         """
+        warnings.warn(
+            'With requests it\'s easy enough for a developer to implement url shortenting themselves. Please see: https://github.com/ryanmcgrath/twython/issues/184',
+            TwythonDeprecationWarning,
+            stacklevel=2
+        )
+
         if shortener == '':
             raise TwythonError('Please provide a URL shortening service.')
 
         request = requests.get(shortener, params={
-            'query': url_to_shorten
+            'format': 'json',
+            'url': url_to_shorten
         })
 
         if request.status_code in [301, 201, 200]:
@@ -349,7 +303,7 @@ class Twython(object):
 
     @staticmethod
     def constructApiURL(base_url, params):
-        return base_url + "?" + "&".join(["%s=%s" % (Twython.unicode2utf8(key), urllib.quote_plus(Twython.unicode2utf8(value))) for (key, value) in params.iteritems()])
+        return base_url + '?' + '&'.join(['%s=%s' % (Twython.unicode2utf8(key), quote_plus(Twython.unicode2utf8(value))) for (key, value) in params.iteritems()])
 
     def searchGen(self, search_query, **kwargs):
         """ Returns a generator of tweets that match a specified query.
@@ -371,37 +325,19 @@ class Twython(object):
         for tweet in content['results']:
             yield tweet
 
-        if 'page' not in kwargs:
-            kwargs['page'] = '2'
-        else:
-            try:
-                kwargs['page'] = int(kwargs['page'])
-                kwargs['page'] += 1
-                kwargs['page'] = str(kwargs['page'])
-            except TypeError:
-                raise TwythonError("searchGen() exited because page takes type str")
+        try:
+            kwargs['page'] = 2 if not 'page' in kwargs else (int(kwargs['page']) + 1)
+        except (TypeError, ValueError):
+            raise TwythonError('Unable to generate next page of search results, `page` is not a number.')
 
         for tweet in self.searchGen(search_query, **kwargs):
             yield tweet
 
-    def bulkUserLookup(self, **kwargs):
-        """Stub for a method that has been deprecated, kept for now to raise errors
-            properly if people are relying on this (which they are...).
-        """
-        warnings.warn(
-            "This function has been deprecated. Please migrate to .lookupUser() - params should be the same.",
-            DeprecationWarning,
-            stacklevel=2
-        )
-
     # The following methods are apart from the other Account methods,
     # because they rely on a whole multipart-data posting function set.
 
     ## Media Uploading functions ##############################################
 
-    def _media_update(self, url, file_, **params):
-        return self.post(url, params=params, files=file_)
-
     def updateProfileBackgroundImage(self, file_, version='1.1', **params):
         """Updates the authenticating user's profile background image.
 
@@ -413,10 +349,11 @@ class Twython(object):
             **params - You may pass items that are stated in this doc
                        (https://dev.twitter.com/docs/api/1.1/post/account/update_profile_background_image)
         """
-        url = 'https://api.twitter.com/%s/account/update_profile_background_image.json' % version
-        return self._media_update(url,
-                                  {'image': (file_, open(file_, 'rb'))},
-                                  **params)
+
+        return self.post('account/update_profile_background_image',
+                         params=params,
+                         files={'image': (file_, open(file_, 'rb'))},
+                         version=version)
 
     def updateProfileImage(self, file_, version='1.1', **params):
         """Updates the authenticating user's profile image (avatar).
@@ -428,10 +365,11 @@ class Twython(object):
             **params - You may pass items that are stated in this doc
                        (https://dev.twitter.com/docs/api/1.1/post/account/update_profile_image)
         """
-        url = 'https://api.twitter.com/%s/account/update_profile_image.json' % version
-        return self._media_update(url,
-                                  {'image': (file_, open(file_, 'rb'))},
-                                  **params)
+
+        return self.post('account/update_profile_image',
+                         params=params,
+                         files={'image': (file_, open(file_, 'rb'))},
+                         version=version)
 
     def updateStatusWithMedia(self, file_, version='1.1', **params):
         """Updates the users status with media
@@ -444,10 +382,10 @@ class Twython(object):
                        (https://dev.twitter.com/docs/api/1.1/post/statuses/update_with_media)
         """
 
-        url = 'https://api.twitter.com/%s/statuses/update_with_media.json' % version
-        return self._media_update(url,
-                                  {'media': (file_, open(file_, 'rb'))},
-                                  **params)
+        return self.post('statuses/update_with_media',
+                         params=params,
+                         files={'media': (file_, open(file_, 'rb'))},
+                         version=version)
 
     def updateProfileBannerImage(self, file_, version='1.1', **params):
         """Updates the users profile banner
@@ -457,23 +395,16 @@ class Twython(object):
                             only API version for Twitter that supports this call
 
             **params - You may pass items that are taken in this doc
-                       (https://dev.twitter.com/docs/api/1/post/account/update_profile_banner)
+                       (https://dev.twitter.com/docs/api/1.1/post/account/update_profile_banner)
         """
-        url = 'https://api.twitter.com/%s/account/update_profile_banner.json' % version
-        return self._media_update(url,
-                                  {'banner': (file_, open(file_, 'rb'))},
-                                  **params)
+
+        return self.post('account/update_profile_banner',
+                         params=params,
+                         files={'banner': (file_, open(file_, 'rb'))},
+                         version=version)
 
     ###########################################################################
 
-    def getProfileImageUrl(self, username, size='normal', version='1'):
-        warnings.warn(
-            "This function has been deprecated. Twitter API v1.1 will not have a dedicated endpoint \
-            for this functionality.",
-            DeprecationWarning,
-            stacklevel=2
-        )
-
     @staticmethod
     def stream(data, callback):
         """A Streaming API endpoint, because requests (by Kenneth Reitz)
